local RayUI, E, L, V, P, G = unpack(select(2, ...))
local L = E.Libs.ACL:GetLocale('ElvUI', E.global.general.locale or 'enUS');
local LSM = E.LSM
local UF = E:GetModule('UnitFrames')

local formattedTextStyles = {
    ['CURRENT'] = '%s',
    ['CURRENT_MAX'] = '%s - %s',
    ['MAX_CURRENT'] = '%s - %s',
    ['CURRENT_PERCENT'] = '%s - %.1f%%',
    ['PERCENT_CURRENT'] = '%.1f%% - %s',
    ['CURRENT_MAX_PERCENT'] = '%s - %s | %.1f%%',
    ['PERCENT_CURRENT_MAX'] = '%.1f%% | %s - %s',
    ['PERCENT'] = '%.1f%%',
    ['DEFICIT'] = '-%s'
}

local _diffcolor = ElvUF.Tags.Methods['difficultycolor']
ElvUF.Tags.Methods['difficultycolor'] = nil
ElvUF.Tags.Methods['difficultycolor'] = function(unit)
    if not UnitIsWildBattlePet(unit) and not UnitIsBattlePetCompanion(unit) and UnitLevel(unit) == -1 then
        local c = QuestDifficultyColors.impossible
        return Hex(c.r, c.g, c.b)
    else
        return _diffcolor(unit)
    end
end

local values = {}
local function GetFormattedText(style, cur, max)
    if max == 0 then max = 1 end

    if style == 'CURRENT' or
        ((style == 'CURRENT_MAX' or style == 'CURRENT_MAX_PERCENT' or style == 'CURRENT_PERCENT' or style == 'PERCENT_CURRENT' or style ==
            'PERCENT_CURRENT_MAX') and cur == max) then
        return format(formattedTextStyles.CURRENT, E:ShortValue(cur))
    else
        local arg1, arg2, arg3 = string.split('_', style)
        local deficit = max - cur
        table.wipe(values)
        values['CURRENT'] = E:ShortValue(cur)
        values['MAX'] = E:ShortValue(max)
        values['PERCENT'] = cur / max * 100
        if cur > 0 and deficit > 0 then
            values['DEFICIT'] = E:ShortValue(deficit)
        else
            values['DEFICIT'] = ''
        end
        return string.format(formattedTextStyles[style], arg1 and values[arg1], arg2 and values[arg2], arg3 and values[arg3],
                             arg4 and values[arg4])
    end
end

for textFormat in pairs(formattedTextStyles) do
    local tagTextFormat = strlower(gsub(textFormat, '_', '-'))
    ElvUF.Tags.Events[format('rayui-health:%s', tagTextFormat)] = 'UNIT_HEALTH UNIT_MAXHEALTH UNIT_CONNECTION PLAYER_FLAGS_CHANGED'
    ElvUF.Tags.Methods[format('rayui-health:%s', tagTextFormat)] = function(unit)
        local status = UnitIsDead(unit) and L['Dead'] or UnitIsGhost(unit) and L['Ghost'] or not UnitIsConnected(unit) and L['Offline']
        if (status) then
            return status
        else
            return GetFormattedText(textFormat, UnitHealth(unit), UnitHealthMax(unit))
        end
    end

    ElvUF.Tags.Events[format('rayui-health:%s-nostatus', tagTextFormat)] = 'UNIT_HEALTH UNIT_MAXHEALTH'
    ElvUF.Tags.Methods[format('rayui-health:%s-nostatus', tagTextFormat)] =
        function(unit)
            return GetFormattedText(textFormat, UnitHealth(unit), UnitHealthMax(unit))
        end

    ElvUF.Tags.Events[format('rayui-power:%s', tagTextFormat)] = 'UNIT_DISPLAYPOWER UNIT_POWER_FREQUENT UNIT_MAXPOWER'
    ElvUF.Tags.Methods[format('rayui-power:%s', tagTextFormat)] = function(unit)
        local pType = UnitPowerType(unit)
        local min = UnitPower(unit, pType)

        return GetFormattedText(textFormat, UnitPower(unit, pType), UnitPowerMax(unit, pType))
    end
end

for _, textFormat in pairs({'veryshort', 'short', 'medium', 'long'}) do
    ElvUF.Tags.Events[format('rayui-health:deficit:name-%s', textFormat)] =
        'UNIT_HEALTH UNIT_MAXHEALTH UNIT_CONNECTION PLAYER_FLAGS_CHANGED UNIT_NAME_UPDATE'
    ElvUF.Tags.Methods[format('rayui-health:deficit:name-%s', textFormat)] =
        function(unit)
            local status = UnitIsDead(unit) and L['Dead'] or UnitIsGhost(unit) and L['Ghost'] or not UnitIsConnected(unit) and L['Offline']
            if (status) then
                return status
            else
                local cur = UnitHealth(unit)
                local max = UnitHealthMax(unit)
                local deficit = max - cur

                if cur > 0 and deficit > 0 then
                    return ElvUF.Tags.Methods['rayui-health:deficit'](unit)
                else
                    return ElvUF.Tags.Methods[format('name:%s', textFormat)](unit)
                end
            end
        end

    ElvUF.Tags.Events[format('rayui-health:deficit-nostatus:name-%s', textFormat)] = 'UNIT_HEALTH UNIT_MAXHEALTH UNIT_NAME_UPDATE'
    ElvUF.Tags.Methods[format('rayui-health:deficit-nostatus:name-%s', textFormat)] =
        function(unit)
            local cur = UnitHealth(unit)
            local max = UnitHealthMax(unit)
            local deficit = max - cur

            if cur > 0 and deficit > 0 then
                return ElvUF.Tags.Methods['rayui-health:deficit'](unit)
            else
                return ElvUF.Tags.Methods[format('name:%s', textFormat)](unit)
            end
        end
end

ElvUF.Tags:RefreshMethods('difficultycolor')
