local RayUI, E, L, V, P, G = unpack(select(2, ...))
local RCT = RayUI:NewModule('RayUICombatText', 'AceEvent-3.0', 'AceHook-3.0', 'AceTimer-3.0', 'AceConsole-3.0')
local xCP = LibStub('xCombatParser-1.0-RayUI', true)

local _G = _G
local pairs, string, type, table, unpack, tonumber, tostring = pairs, string, type, table, unpack, tonumber, tostring
local ipairs, math, next = ipairs, math, next
local CreateFrame = CreateFrame
local SetCVar = SetCVar
local GetSpellTexture = GetSpellTexture
local GetSpellInfo = GetSpellInfo

local showreceived = true
local showoutput = true

local showdots = true
local showhots = true

local ctshowpet = true

local unmergedEvents = {}
local mergedEvents = {}

-- local lastMergeUpdate = 0
-- local lastThrottleUpdate = 0

local frameIndex = {[1] = 'outgoing', [2] = 'incomingDamage', [3] = 'incomingHeal', [4] = 'message', [5] = 'power'}

local createdFrames = {}

local spellColors = {
    -- Vanilla Schools
    [SCHOOL_MASK_PHYSICAL] = {1.00, 1.00, 1.00},
    [SCHOOL_MASK_HOLY] = {1.00, 0.90, 0.50},
    [SCHOOL_MASK_FIRE] = {1.00, 0.5, 0},
    [SCHOOL_MASK_NATURE] = {0.30, 1.00, 0.30},
    [SCHOOL_MASK_FROST] = {0.5, 1.00, 1.00},
    [SCHOOL_MASK_SHADOW] = {0.50, 0.50, 1.00},
    [SCHOOL_MASK_ARCANE] = {1.00, 0.50, 1.00},

    -- Physical and a Magical
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FIRE] = {1.00, 0.58, 0.59},
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FROST] = {0.65, 0.65, 0.95},
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_ARCANE] = {0.87, 0.87, 0.87},
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_NATURE] = {0.70, 1.00, 0.70},
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_SHADOW] = {1.00, 0.85, 1.00},
    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_HOLY] = {1.00, 1.00, 0.83},

    -- Two Magical Schools
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST] = {0.65, 0.23, 0.54},
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_ARCANE] = {0.87, 0.45, 0.47},
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_NATURE] = {0.70, 0.58, 0.29},
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_SHADOW] = {1.00, 0.43, 0.59},
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_HOLY] = {1.00, 0.58, 0.24},
    [SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE] = {0.53, 0.53, 0.83},
    [SCHOOL_MASK_FROST + SCHOOL_MASK_NATURE] = {0.35, 0.65, 0.65},
    [SCHOOL_MASK_FROST + SCHOOL_MASK_SHADOW] = {0.65, 0.50, 0.95},
    [SCHOOL_MASK_FROST + SCHOOL_MASK_HOLY] = {0.65, 0.65, 0.60},
    [SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE] = {0.58, 0.87, 0.58},
    [SCHOOL_MASK_ARCANE + SCHOOL_MASK_SHADOW] = {0.87, 0.73, 0.87},
    [SCHOOL_MASK_ARCANE + SCHOOL_MASK_HOLY] = {0.87, 0.87, 0.53},
    [SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW] = {0.70, 0.85, 0.70},
    [SCHOOL_MASK_NATURE + SCHOOL_MASK_HOLY] = {0.70, 1.00, 0.35},
    [SCHOOL_MASK_SHADOW + SCHOOL_MASK_HOLY] = {1.00, 0.85, 0.65},

    -- Three or More Schools
    [SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_NATURE] = {0.57, 0.48, 0.49},

    [SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW] = {0.69, 0.58, 0.65},

    [SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW + SCHOOL_MASK_HOLY] = {
        0.74, 0.65, 0.59
    },

    [SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST + SCHOOL_MASK_ARCANE + SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW +
        SCHOOL_MASK_HOLY] = {0.78, 0.70, 0.65},

    ['HEAL'] = {0, 1.00, 0},
    ['MISS'] = {0.50, 0.50, 0.50},
    ['DISPEL_BUFF'] = {0, 1.00, 0.50},
    ['INTERRUPT'] = {1.00, 0, 0.50}
}

local enablePower = {
    ['MANA'] = true,
    ['RAGE'] = true,
    ['FOCUS'] = true,
    ['ENERGY'] = true,

    ['RUNES'] = false,
    ['RUNIC_POWER'] = true,
    ['SOUL_SHARDS'] = true,
    ['LUNAR_POWER'] = false,

    ['HOLY_POWER'] = false,
    ['ALTERNATE_POWER'] = false,
    ['CHI'] = false,
    ['INSANITY'] = true,

    ['ARCANE_CHARGES'] = false,
    ['FURY'] = true,
    ['PAIN'] = true
}

local format_spell_icon = ' |T%s:%d:%d:0:0:64:64:5:59:5:59|t'

local function EraseTable(t)
    -- Loop through all the keys in the table and clear it.
    for key in next, t do t[key] = nil end
end

function RCT:GetSpellTextureFormatted(spellID, message, critical, mergeTrailer, iconSize, justifyH)
    iconSize = critical and (iconSize + 10) or iconSize
    local icon = ''
    if spellID == 0 then
        icon = _G.PET_ATTACK_TEXTURE
    elseif type(spellID) == 'string' then
        icon = spellID
    else
        icon = GetSpellTexture(spellID) or ''
    end
    if not mergeTrailer then mergeTrailer = '' end

    if icon ~= '' then
        if justifyH == 'LEFT' or justifyH == 'CENTER' then
            message = string.format(critical and '%s |cffff0000*|r%s|cffff0000*|r%s' or '%s %s%s',
                                    string.format(format_spell_icon, icon, iconSize, iconSize), message, mergeTrailer)
        else
            message = string.format(critical and '|cffff0000*|r%s|cffff0000*|r%s%s' or '%s%s%s', message, mergeTrailer,
                                    string.format(format_spell_icon, icon, iconSize, iconSize))
        end
    else
        message = string.format(critical and '|cffff0000*|r%s|cffff0000*|r%%s' or '%s%s', message, mergeTrailer)
    end

    return message
end

function RCT:AddMessage(framename, message, color)
    self.frames[framename]:AddMessage(message, unpack(color))
end

local index = 1
local frames = {}

-- code from MSBT
local function MergeEvents(numEvents)
    -- Holds an unmerged event and whether or not to merge it.
    local unmergedEvent
    local doMerge = false

    -- Don't attempt to merge any more events than were available when the function was called since
    -- more events may get added while the merge is taking place.
    for i = 1, numEvents do
        -- Get the unmerged event.
        unmergedEvent = unmergedEvents[i]

        -- Loop through all of the events in the merged events array.
        for _, mergedEvent in ipairs(mergedEvents) do
            -- Check if the event types match.
            if (unmergedEvent.event == mergedEvent.event) then
                -- Check if there is no skill name.
                if (not unmergedEvent.spellName) then
                    -- Set the merge flag if the affected unit name is the same.
                    if ((unmergedEvent.sourceName == mergedEvent.sourceName) and unmergedEvent.sourceName) then
                        doMerge = true
                    end

                    -- The skill names match.
                elseif (unmergedEvent.spellName == mergedEvent.spellName) then
                    -- Change the name to the multiple targets string if the names don't match.
                    if (unmergedEvent.sourceName ~= mergedEvent.sourceName) then
                        mergedEvent.sourceName = L.MSG_MULTIPLE_TARGETS
                    end

                    -- Set the merge flag.
                    doMerge = true
                end
            end -- Event types match.

            -- Check if the event should be merged.
            if (doMerge) then
                -- Set the event merged flag for the event being merged.
                unmergedEvent.eventMerged = true

                -- Total the amount if there is one.
                if (unmergedEvent.amount) then mergedEvent.amount = (mergedEvent.amount or 0) + unmergedEvent.amount end

                -- Increment the number of merged events.
                mergedEvent.numMerged = mergedEvent.numMerged + 1

                -- Increment the number of crits if the event being merged is a crit. Clear the crit flag for the merged event if it isn't.
                if (unmergedEvent.critical) then
                    mergedEvent.numCrits = mergedEvent.numCrits + 1
                else
                    mergedEvent.critical = false
                end

                -- Break out of the merged events loop since the event has been merged.
                break
            end -- Do Merge.
        end -- Loop through merged events.

        -- Add the event to the end of the merged events array if it wasn't merged.
        if (not doMerge) then
            unmergedEvent.numMerged = 0

            -- Set the number of crits depending on if the event is a crit or not.
            if (unmergedEvent.critical) then
                unmergedEvent.numCrits = 1
            else
                unmergedEvent.numCrits = 0
            end

            -- Add the event to the end of the merged events array.
            mergedEvents[#mergedEvents + 1] = unmergedEvent
        end

        -- Reset the event merge flag.
        doMerge = false
    end -- Loop through unmerged events.

    -- Append merge trailer information to the merged events if enabled.
    for _, mergedEvent in ipairs(mergedEvents) do
        -- Check if there were any events merged.
        if (mergedEvent.numMerged > 0) then
            -- Create the crit trailer text if there were any crits.
            local critTrailer = ''
            if (mergedEvent.numCrits > 0) then
                critTrailer = string.format('(|cffff0000%d|r)', mergedEvent.numCrits)
                mergedEvent.critical = true
            end

            -- Set the event's merge trailer field.
            mergedEvent.mergeTrailer = string.format(' |cffffffffx%d%s|r', mergedEvent.numMerged + 1, critTrailer)
        end -- Events were merged.
    end -- Loop through merged events.

    -- Remove the processed events from unmerged events queue.
    for i = 1, numEvents do
        -- Recycle the unmerged event if it was merged.
        if (unmergedEvents[1].eventMerged) then EraseTable(unmergedEvents[1]) end

        -- Remove the event from the unmerged events array.
        table.remove(unmergedEvents, 1)
    end
end

function RCT:OnSpamUpdate()
    MergeEvents(#unmergedEvents)

    -- Display and recycle the merged events.
    local eventSettings, hideSkills, outputMessage
    for i, combatEvent in ipairs(mergedEvents) do
        -- Is the source someone we care about?
        if combatEvent.isPlayer or combatEvent:IsSourceMyVehicle() or ctshowpet and combatEvent:IsSourceMyPet() then
            if combatEvent.suffix == '_HEAL' then
                self:HealingOutgoing(combatEvent)

            elseif combatEvent.suffix == '_DAMAGE' then
                self:DamageOutgoing(combatEvent)

            elseif combatEvent.suffix == '_MISSED' then
                self:OutgoingMiss(combatEvent)

            elseif combatEvent.event == 'PARTY_KILL' then
                -- self:KilledUnit(combatEvent)

            elseif combatEvent.event == 'SPELL_INTERRUPT' then
                self:InterruptedUnit(combatEvent)

            elseif combatEvent.event == 'SPELL_DISPEL' then
                self:SpellDispel(combatEvent)

            elseif combatEvent.event == 'SPELL_STOLEN' then
                self:SpellStolen(combatEvent)

            end
        end

        -- Is the destination someone we care about?
        if combatEvent.atPlayer or combatEvent:IsDestinationMyVehicle() then
            if combatEvent.suffix == '_HEAL' then
                self:HealingIncoming(combatEvent)

            elseif combatEvent.suffix == '_DAMAGE' then
                self:DamageIncoming(combatEvent)

            elseif combatEvent.suffix == '_MISSED' then
                self:IncomingMiss(combatEvent)

            end
        end
        mergedEvents[i] = nil
        EraseTable(combatEvent)
    end
end

function RCT:HealingOutgoing(args)
    local critical, spellID, isHoT, amount, mergeTrailer = args.critical, args.spellId, args.prefix == 'SPELL_PERIODIC', args.amount,
                                                           args.mergeTrailer
    if isHoT and not showhots then return end
    local message = RCT:GetSpellTextureFormatted(spellID, E:ShortValue(amount), critical, mergeTrailer, self.db.iconSize,
                                                 self.db.outgoing.justifyH)
    RCT:AddMessage('outgoing', message, spellColors['HEAL'])
end

function RCT:HealingIncoming(args)
    local critical, spellID, isHoT, amount, mergeTrailer = args.critical, args.spellId, args.prefix == 'SPELL_PERIODIC', args.amount,
                                                           args.mergeTrailer
    if isHoT and not showhots then return end
    local message = RCT:GetSpellTextureFormatted(spellID, '+' .. E:ShortValue(amount), critical, mergeTrailer, self.db.iconSize,
                                                 self.db.incomingHeal.justifyH)
    RCT:AddMessage('incomingHeal', message, spellColors['HEAL'])
end

function RCT:DamageOutgoing(args)
    local critical, spellID, amount, mergeTrailer = args.critical, args.spellId, args.amount, args.mergeTrailer
    local isSwing = args.prefix == 'SWING'
    if isSwing and (args:IsSourceMyPet() or args:IsSourceMyVehicle()) then spellID = 0 end
    local message = RCT:GetSpellTextureFormatted(spellID, E:ShortValue(amount), critical, mergeTrailer, self.db.iconSize,
                                                 self.db.outgoing.justifyH)
    RCT:AddMessage('outgoing', message, spellColors[args.spellSchool])
end

function RCT:DamageIncoming(args)
    local critical, amount, spellID, mergeTrailer = args.critical, args.amount, args.spellId, args.mergeTrailer
    local isSwing = args.prefix == 'SWING'
    if isSwing then spellID = 0 end
    local message = RCT:GetSpellTextureFormatted(spellID, '-' .. E:ShortValue(amount), critical, mergeTrailer, self.db.iconSize,
                                                 self.db.incomingDamage.justifyH)
    RCT:AddMessage('incomingDamage', message, spellColors[args.spellSchool])
end

function RCT:OutgoingMiss(args)
    local message, spellId, mergeTrailer = _G['COMBAT_TEXT_' .. args.missType], args.spellId, args.mergeTrailer
    message = RCT:GetSpellTextureFormatted(spellId, message, nil, mergeTrailer, self.db.iconSize, self.db.outgoing.justifyH)
    RCT:AddMessage('outgoing', message, spellColors['MISS'])
end

function RCT:IncomingMiss(args)
    local message, spellId, mergeTrailer = _G['COMBAT_TEXT_' .. args.missType], args.extraSpellId, args.mergeTrailer
    message = RCT:GetSpellTextureFormatted(spellId, message, nil, mergeTrailer, self.db.iconSize, self.db.incomingDamage.justifyH)
    RCT:AddMessage('incomingDamage', message, spellColors['MISS'])
end

function RCT:InterruptedUnit(args)
    local message = RCT:GetSpellTextureFormatted(args.extraSpellId, args.extraSpellName, nil, nil, self.db.iconSize - 3,
                                                 self.db.message.justifyH)
    RCT:AddMessage('message', string.format('%s: %s', _G['ACTION_' .. args.event], message), spellColors['INTERRUPT'])
end

function RCT:SpellDispel(args)
    local message = RCT:GetSpellTextureFormatted(args.extraSpellId, args.extraSpellName, nil, nil, self.db.iconSize - 3,
                                                 self.db.message.justifyH)
    RCT:AddMessage('message', string.format('%s: %s', _G['ACTION_' .. args.event], message),
                   args.auraType == 'BUFF' and spellColors['DISPEL_BUFF'] or spellColors['INTERRUPT'])
end

function RCT:SpellStolen(args)
    local message = RCT:GetSpellTextureFormatted(args.extraSpellId, args.extraSpellName, nil, nil, self.db.iconSize - 3,
                                                 self.db.message.justifyH)
    RCT:AddMessage('message', string.format('%s: %s', _G['ACTION_' .. args.event], message),
                   args.auraType == 'BUFF' and spellColors['DISPEL_BUFF'] or spellColors['INTERRUPT'])
end

function RCT.CombatLogEvent(args)
    if args.sourceGUID == args.destGUID then args.event = 'SELF_' .. args.event end
    unmergedEvents[#unmergedEvents + 1] = {}
    E:CopyTable(unmergedEvents[#unmergedEvents], args)
end

function RCT:COMBAT_TEXT_UPDATE(event, subevent, ...)
    if subevent == 'ENERGIZE' or subevent == 'PERIODIC_ENERGIZE' then
        local amount, energy_type = ...
        if enablePower[energy_type] then
            RCT:AddMessage('power', string.format('+%s %s', E:ShortValue(amount), _G[energy_type]), _G.ElvUF.colors.power[energy_type])
        end
    end
end

function RCT:CreateMessageFrame(name)
    local f = CreateFrame('ScrollingMessageFrame', 'RayUI_CombatText_' .. string.gsub(name, '%s+', ''), E.UIParent)

    f:CreateBackdrop('Transparent')
    f.title = f:CreateFontString('OVERLAY')
    f.title:FontTemplate()
    f.title:Point('CENTER', f, 'CENTER')
    f.title:SetText(L[name])
    f.name = name

    f.backdrop:Hide()
    f.title:Hide()

    createdFrames[f] = true
    return f
end

function RCT:Configure_CombatTextFrame(ftype)
    local frame = self.frames[ftype]
    local db = self.db[ftype]

    frame:Clear()
    frame:SetFadeDuration(0.2)
    frame:SetSpacing(2)
    frame:SetWidth(db.width)
    frame:SetHeight(db.height)
    frame:SetJustifyH(db.justifyH)
    frame.justifyH = db.justifyH
    frame:SetFont(E.Libs.LSM:Fetch('font', self.db.font), self.db.fontSize, self.db.fontOutline)
    frame:SetTimeVisible(self.db.fadeTime)
    frame:SetMaxLines(db.height / (self.db.iconSize + 10))
end

function RCT:Configure_CombatText()
    for ftype in pairs(self.frames) do self:Configure_CombatTextFrame(ftype) end
    if self.timer then
        self:CancelTimer(self.timer);
        self.timer = nil
    end
    self.timer = self:ScheduleRepeatingTimer('OnSpamUpdate', self.db.mergeDelay)
end

do
    local missType = {[1] = 'DODGE', [2] = 'MISS', [3] = 'PARRY', [4] = 'BLOCK'}
    local handlers = {
        [1] = 'HealingOutgoing',
        [2] = 'DamageOutgoing',
        [3] = 'OutgoingMiss',
        [4] = 'InterruptedUnit',
        [5] = 'SpellDispel',
        [6] = 'SpellStolen',
        [7] = 'HealingIncoming',
        [8] = 'HealingIncoming',
        [9] = 'DamageIncoming',
        [10] = 'HealingIncoming',
        [11] = 'IncomingMiss',
        [12] = 'DamageIncoming'
    }
    local events = {[1] = 'PARTY_KILL', [2] = 'SPELL_INTERRUPT', [3] = 'SPELL_DISPEL', [4] = 'SPELL_STOLEN'}
    local current = 1
    function RCT:ForceShow_Update()
        local event = {}
        event.critical = math.random(1, 2) == 1 and true or false
        event.amount = math.random() * 1e5
        while event.spellName == nil do
            event.spellId = math.random(100, 10000)
            event.spellName = GetSpellInfo(event.spellId)
        end
        event.extraSpellId = event.spellId
        event.extraSpellName = event.spellName
        while spellColors[event.spellSchool or 999] == nil do event.spellSchool = math.random(1, 127) end
        event.missType = missType[math.random(1, 4)]
        event.event = events[math.random(1, 4)]

        if not handlers[current] then current = 1 end
        self[handlers[current]](self, event)
        current = current + 1
    end
end

function RCT:ToggleForceShowMode()
    if self.isForced then
        for ftype, frame in pairs(self.frames) do
            frame:Clear()
            frame.backdrop:Hide()
            frame.title:Hide()
        end
        self:CancelTimer(self.forceTimer)
        self.forceTimer = nil
        self.isForced = false
    else
        for ftype, frame in pairs(self.frames) do
            frame:Clear()
            frame.backdrop:Show()
            frame.title:Show()
        end
        self.forceTimer = self:ScheduleRepeatingTimer('ForceShow_Update', 0.2)
        self.isForced = true
    end
end

function RCT:Initialize()
    if not E.db.rayui.combattext.enable then return end

    RayUI:RegisterDB(self, 'combattext')

    self.db = E.db.rayui.combattext
    self.frames = {}
    self.frames['outgoing'] = self:CreateMessageFrame('Outgoing')
    self.frames['incomingDamage'] = self:CreateMessageFrame('Incoming Damage')
    self.frames['incomingHeal'] = self:CreateMessageFrame('Incoming Healing')
    self.frames['message'] = self:CreateMessageFrame('Message')

    self:Configure_CombatText()

    self.frames['outgoing']:SetPoint('BOTTOMRIGHT', 'ElvUF_TargetTarget', 'TOPRIGHT', 0, 20)
    self.frames['incomingDamage']:SetPoint('BOTTOMLEFT', 'ElvUF_Player', 'TOPLEFT', -150, 20)
    self.frames['incomingHeal']:SetPoint('BOTTOMLEFT', 'ElvUF_Player', 'TOPLEFT', -250, 20)
    self.frames['message']:SetPoint('BOTTOM', E.UIParent, 'CENTER', 0, 200)

    for ftype, frame in pairs(self.frames) do
        E:CreateMover(frame, 'CombatText ' .. frame.name .. ' Mover', L['Combat Text'] .. ' - ' .. L[frame.name], true, nil, nil,
                      'ALL,GENERAL,RAYUI')
    end

    xCP:RegisterCombat(self.CombatLogEvent)

    self:RegisterEvent('COMBAT_TEXT_UPDATE')

    if _G.CombatText then
        _G.CombatText:UnregisterAllEvents()
        _G.CombatText:SetScript('OnLoad', nil)
        _G.CombatText:SetScript('OnEvent', nil)
        _G.CombatText:SetScript('OnUpdate', nil)
    end

    _G.InterfaceOptionsCombatPanelEnableFloatingCombatText:Disable()
    _G.InterfaceOptionsCombatPanelEnableFloatingCombatTextText:SetTextColor(.5, .5, .5)

    SetCVar('enableFloatingCombatText', 0)
end

RayUI:RegisterModule(RCT:GetName())
